FUNCTION PAYTYPE
* This function check the payment type as a function
* of the total order amount to validate the record.
LOCAL llReturnCode, lnChangeCheck

* Check if any changes were made to either the sales total or
* the payment method.
* Check if customer attempts to pay < $10 by credit card.
  IF (MOD(GETFLDSTATE('nSalesTotal'),2) = 0 OR ;
      MOD(GETFLDSTATE('cPaymentMethod'),2) = 0) AND ;
      nSalesTotal < 10.00
  * Check for payment method of 'CA' - CASH
    IF cPaymentMethod # 'CA'
      = MESSAGEBOX('Orders of less than $10 must be cash')
      RETURN .F.
    ENDIF
  ENDIF

* If paid by credit card, check if credit card was approved.
  IF (MOD(GETFLDSTATE('nSalesTotal'),2) = 0 OR ;
      MOD(GETFLDSTATE('cPaymentMethod'),2) = 0) AND ;
      cPaymentMethod # 'CA'
  * Ask if card was approved.  If not reject the record.
    IF MESSAGEBOX('Was card approved?', 36) = 7
      = MESSAGEBOX('Cannot accept a credit card order without approval')
      RETURN .F.
    ENDIF
  ENDIF

RETURN .T.


****************
FUNCTION RISALES
* This routine is called by the triggers in SALES.DBF
LPARAMETERS lAction
LOCAL llReturnValue, Dummy

* Start TranslAction
  Dummy         = IIF(TYPE("nLevel")<>"N" OR nLevel=0, ;
                  RISTART(), "")
  nLevel        = nLevel + 1
  llReturnValue = .F.

* Perform lAction
  DO CASE
    CASE TYPE('lAction') # 'C' ;
         OR !UPPER(lAction) $ 'DIU' ;
         OR LEN(lAction)>1
      ERROR 'Invalid lAction code passed to RISALES'
    CASE UPPER(lAction) == 'D'
      llReturnValue = SALEDEL()
    CASE UPPER(lAction) == 'I'
      llReturnValue = SALEINS()
    CASE UPPER(lAction) == 'U'
      llReturnValue = SALEMOD()
  ENDCASE

* End translAction
  nLevel = nLevel - 1
  Dummy   = IIF(nLevel = 0, RIEND(llReturnValue), 0)
RETURN llReturnValue


****************
FUNCTION SALEDEL
* If deleting a sales record, delete all details
LOCAL  llReturnValue, lcOrderId, lnDetArea

  llReturnValue = .T.
  IF nLevel = 1
    lcOrderId     = cOrderId
    lnDetArea     = RIOPEN('DETAILS')
    SELECT (lnDetArea)
    SCAN FOR cOrderId = lcOrderid
      llReturnValue = RIDELETE()
    ENDSCAN
    SELECT sales
  ENDIF

RETURN llReturnValue


****************
FUNCTION SALEINS
* No referential lAction required when adding a parent
RETURN .T.


****************
FUNCTION SALEMOD
* If modifying a sales record, modify all details
LOCAL  llReturnValue, lcOrderId, lcOldValue, lnDetArea

  llReturnValue = .T.
  IF nLevel = 1
    lcOrderId     = cOrderId
    lcOldValue    = OLDVAL('cOrderId')

  * If key value changed, updated the child records
    IF lcOrderId <> lcOldValue
      lnDetArea     = RIOPEN('DETAILS')
      SELECT (lnDetArea)
      SCAN FOR cOrderId = lcOldValue
        REPLACE cOrderId WITH lcOrderId
      ENDSCAN
      SELECT sales
    ENDIF
  ENDIF
RETURN llReturnValue


******************
FUNCTION RIDETAILS
* This routine is called by the triggers in DETAILS.DBF
LPARAMETERS lAction
LOCAL llReturnValue, Dummy

* Start TranslAction
  Dummy          = IIF(TYPE("nLevel")<>"N" OR nLevel=0, ;
                   RISTART(), "")
  nLevel        = nLevel + 1
  llReturnValue = .F.

* Perform lAction
  DO CASE
    CASE TYPE('lAction') # 'C' ;
         OR !UPPER(lAction) $ 'DIU' ;
         OR LEN(lAction)>1
      ERROR 'Invalid lAction code passed to RIDETAILS'
    CASE UPPER(lAction) == 'D'
      llReturnValue = DETAILDEL()
    CASE UPPER(lAction) == 'I'
      llReturnValue = DETAILINS()
    CASE UPPER(lAction) == 'U'
      llReturnValue = DETAILMOD()
  ENDCASE

* End translAction
  nLevel = nLevel - 1
  Dummy   = IIF(nLevel = 0, RIEND(llReturnValue), 0)
RETURN llReturnValue


******************
FUNCTION DETAILDEL
* Check if all order details are deleted, then delete parent
LOCAL cnt, lcOrderId, lnCurrentArea, ;
      lnSalesArea, lnDetArea, llReturnValue

  llReturnValue   = .T.
  lcOrderId       = cOrderId
  lnCurrentArea   = SELECT()

  IF !EMPTY(lcOrderId)
    lnDetArea     = RIOPEN('DETAILS')
    cnt           = 0
    SELECT (lnDetArea)
    SCAN FOR cOrderId = lcOrderId AND !DELETED()
      cnt         = cnt + 1
    ENDSCAN
    = RIREUSE('DETAILS', lnDetArea)
  
    IF cnt = 1 AND nLevel = 1
      lnSaleArea = RIOPEN('SALES', 'ORDERID')
      IF SEEK(lcOrderId)
        llReturnValue = RIDELETE()
      ENDIF
      = RIREUSE('SALES', lnSaleArea)
    ENDIF
    SELECT (lnCurrentArea)
  ENDIF
RETURN llReturnValue


******************
FUNCTION DETAILINS
* Insert a detail record only if a sales record exists
LOCAL lcOrderId, lnSaleArea, llReturnValue, lnCurrentArea

  llReturnValue =  .T.
  lcOrderId     = cOrderId
  lnCurrentArea = SELECT()

  IF !EMPTY(lcOrderId)
    lnSaleArea    = RIOPEN('SALES', 'ORDERID')
    llReturnValue = SEEK(lcOrderId, lnSaleArea) AND ;
                     !DELETED(lnSaleArea)
    = RIREUSE('SALES', lnSaleArea)
    IF !llReturnValue
      = MESSAGEBOX('No corresponding order for this detail')
    ENDIF
    SELECT (lnCurrentArea)
  ENDIF
RETURN llReturnValue


******************
FUNCTION DETAILMOD
* Allow key value change if it matches another parent
* If no more child records for this parent, delete it
LOCAL cnt, lcOrderId, lcOldValue, ;
      lnSalesArea, llReturnValue, lnDetArea

  llReturnValue   = .T.

  IF nLevel=1
    lcOrderId       = cOrderId
    lcOldValue      = OLDVAL('cOrderId')

  * First check if new value is a valid parent key
    lnSaleArea      = RIOPEN('SALES', 'ORDERID')
    llReturnValue   = SEEK(lcOrderId, lnSaleArea) AND ;
                       !DELETED(lnSaleArea)
    IF !llReturnValue
      = MESSAGEBOX('No corresponding order for this detail')
    ENDIF

  * New order id is valid, check for other child records in old order
    IF llReturnValue
      cnt              = 0
      lnDetArea       = RIOPEN('DETAILS')
      SELECT (lnDetArea)
      SCAN FOR OLDVAL('cOrderId') = lcOldValue AND !DELETED()
        cnt = cnt + 1
      ENDSCAN
      = RIREUSE('DETAILS', lnDetArea)

    * If no other child records, delete the parent
      IF cnt = 1
        SELECT (lnSaleArea)
        IF SEEK(lcOldValue)
          llReturnValue = RIDELETE()
        ENDIF
      ENDIF
    ENDIF
    SELECT DETAILS
  ENDIF
  = RIREUSE('SALES', lnSaleArea)
RETURN llReturnValue


*****************
FUNCTION RIDELETE
* Delete the current record in the current area
LOCAL llReturnValue

llReturnValue = .T.
* Attempt to get a record lock
  IF (UPPER(SYS(2011))='RECORD LOCKED' and !DELETED()) OR !RLOCK()
    llReturnValue = .F.
  ELSE
  * If not deleted, delete it.
    IF !DELETED()
      DELETE
      nError = 0
      UNLOCK RECORD (RECNO())
      IF nError <> 0
        = TABLEREVERT()
        UNLOCK RECORD (RECNO())
        llReturnValue = .F.
      ENDIF
    ENDIF
  ENDIF
RETURN llReturnValue


****************
FUNCTION RISTART
* Call this program when starting Referential Integrity.
* It initializes several variables.
  IF TYPE("nLevel") <> 'N'
    PUBLIC nLevel
    nLevel = 0
  ENDIF

  IF TYPE("cCursors") <> "C"
    PUBLIC cCursors
  ENDIF

  IF nLevel = 0
    BEGIN TRANSACTION
    PUBLIC cOldError, nError
    cCursors  = ""
    cOldError = ON("ERROR")
    nError    = 0
    ON ERROR nError = ERROR()
  ENDIF
RETURN


***************
FUNCTION RIOPEN
* This procedure opens cursors for use during Referential
* Integrity checks since operations that perform record
* pointer moves are not allowed directly.
PARAMETERS cTable, cTag
LOCAL lnNewArea, nInUseArea

  nInUseArea = ATC(cTable+"*", cCursors)
  
* Open reference table
  IF nInUseArea= 0
    SELECT 0
    lnNewArea = SELECT()
    IF EMPTY(cTag)
      USE (cTable) ;
          ALIAS ("RI_"+LTRIM(STR(lnNewArea))) ;
          AGAIN SHARE
    ELSE
      USE (cTable) ;
          ORDER (cTag) ;
          ALIAS ("RI_"+LTRIM(STR(lnNewArea))) ;
          AGAIN SHARE
    ENDIF
    cCursors = cCursors + UPPER(cTable) + "?" + ;
                STR(lnNewArea,5)
  ELSE
  * Retrieve work area of referential integrity cursor
    lnNewArea = VAL(SUBSTR(cCursors, ;
                nInUseArea + LEN(cTable) + 1, 5))
    cCursors  = STRTRAN(cCursors, ;
                UPPER(cTable) + "*" + STR(lnNewArea,5), ;
                UPPER(cTable) + "?" + STR(lnNewArea,5))
  ENDIF

RETURN (lnNewArea)


**************
FUNCTION RIEND
* Call this routine to exit the referential integrity check
* It saves changes or reverts to original values depending
* on the value passed to it.
LPARAMETER lSuccess
LOCAL lnXx

* Complete translAction or roll it back
  IF !lSuccess
    IF USED('SALES')
      = TABLEREVERT(.T., 'SALES')
    ENDIF
    IF USED('DETAILS')
      = TABLEREVERT(.T., 'DETAILS')
    ENDIF
  ENDIF
  END TRANSACTION
*  IF lSuccess
*    END TRANSACTION
*  ELSE
*    ROLLBACK
*  ENDIF

* Reset on error
  IF EMPTY(cOldError)
    ON ERROR
  ELSE
    ON ERROR (cOldError)
  ENDIF

* Remove cursors and reset variables that track them
  FOR lnXx = 1 TO OCCURS("?", cCursors)
    cFound = ATC('?', cCursors, lnXx) + 1
    USE IN (VAL(SUBSTR(cCursors, cFound, 5)))
  ENDFOR
  FOR lnXx = 1 TO OCCURS("*", cCursors)
    cFound = ATC('*', cCursors, lnXx) + 1
    USE IN (VAL(SUBSTR(cCursors, cFound, 5)))
  ENDFOR
  STORE "" TO cCursors
RETURN .T.


****************
FUNCTION RIREUSE
* This routine allows reuse of exiting cursor
PARAMETERS cTable, cArea
  cCursors = STRTRAN(cCursors, ;
              UPPER(cTable) + "?" + STR(cArea,5), ;
              UPPER(cTable) + "*" + STR(cArea,5))
RETURN .T.
